#ifndef CMA_HEAP_DEFINED
#define CMA_HEAP_DEFINED

#include <stdlib.h>
#include <assert.h>

namespace CMa
{
    /**
     * An parametrized implementation of CMa heap. This implementation does not allocate memory on demand.
     *
     *  Three important C-machine registers must always satisfy inequalities:
     *  - \f$0\le SP\le EP<NP<MAX\f$
     *  - \f$0\le FP<SP\f$
     *
     *  Heap methods verifytXXX have following semantics:
     *   - if everything is OK   they return TRUE
     *   - if something is WRONG they return FALSE
     *   - they dont change state of Heap
     *
     *  There are only following test methods:
     *  - verifySP, verifyNP, verifyEP, verifyFP
     *  - verifyAddress(validates address in the heap)
     *  - verifyInterval(validates address interval in te heap)
     *
     *  Heap methods setXXX,[],top() DON'T test validity of parameters. The validity must be assured
     *  before by verifyXXX calls. Set methods return nothing(void).
     *
     *  There are only following set methods:
     *  - setSP
     *  - setNP
     *  - setEP
     *  - setFP
     *
     *  The values in the interval [SP+1, NP-1] may CHANGE any time.
     *
     *  @author OM/Modified by SL
     *
     *  STATE: NOT TESTED
     *
     *
     */
	template<class Word, class Int> class Heap
	{
	protected:

		Word* data;
		int MAX;

		Int SP;
		Int NP;
		Int EP;

	public:

		Heap()
		{
			data = NULL;
			MAX = 0;
		}

		/**
		 * Sets the parameters for the heap.
		 *
		 * @param virtualSize	the size of heap as percieved by the program
		 * @param realSize		amount of memory allocated for the heap
		 *
		 * Assumptions:
		 *		- realSize > 0
		 *		- virtualSize >= realSize
		 *
		 * @author OM
		 */
		void setParams(Int virtualSize, int realSize)
		{
			assert(realSize > 0);
			assert(virtualSize >= (Int)realSize);
			MAX = realSize;
		}



		/* Interfaces used by the templates */

		/** Initializes the heap. May fail */
		void init() throw (Failure)
		{
			data = new Word[MAX];
			if (data == NULL)
				throw Failure("Not enough memory");

			SP = (Int)-1;
			NP = (Int)MAX;
			EP = (Int)0;
		}

		void deInit()
		{
			delete data;
			data = NULL;
		}

		~Heap()
		{
			if (data != NULL) delete data;
		}

		/** Get the size of the heap. */
		inline Int getSize()
		{
			return MAX;
		}
        /******************************************Test*methods******************************************/
		inline bool verifySP(const Int& newSP)
		{
			return (Int(0) <= newSP && newSP <= EP );
		}
		
		inline bool verifyNP(const Int& newNP)
		{
			return (EP < newNP && newNP <(Int)MAX);
		}

		inline bool verifyEP(const Int& newEP)
		{
			return (SP <= newEP && newEP < NP);
		}
						
		inline bool verifyFP(const Int& newFP, const Int& newSP)
		{
			return ((Int)0<= newFP && newFP < newSP);
		}
		
		inline bool verifyAddress(const Int& addr)
		{
			return (((Int)0 <= addr &&  addr<=SP) || (NP<=addr && addr < (Int)MAX));
		}

		inline bool verifyInterval(const Int& addr1, const Int& addr2)
		{
			//interval is invalid
			if(addr1> addr2) return false;
			//Interval is in the stack
			if(addr1 <= SP) return ((Int)0 <= addr1 && addr2 <= SP);
			//Interval is in the heap
			return (NP<=addr1 && addr2 < (Int)MAX);
		}
		
        /*********************************************Get*methods********************************************/
		inline Int getSP() { return SP; }
		inline Int getNP() { return NP; }
		inline Int getEP() { return EP; }

		/*********************************************Set*Methods********************************************/
		/*                                     NO VALIDITY CHECKS ARE DONE                                  */
		/****************************************************************************************************/
		inline void setSP(const Int& newSP)
		{
			assert(verifySP(newSP));
			SP = newSP;
		}
		
		inline void setNP(const Int& newNP)
		{
			assert(verifyNP(newNP));
			NP = newNP;
		}

		inline void setEP(const Int& newEP)
		{
			assert(verifyEP(newEP));
			EP = newEP;
		}


		/*********************************************Heap*access********************************************/
		/*                            ALL VALIDITY CHECKS MUST BE DONE BEFORE                               */
		/****************************************************************************************************/
		inline Word& operator[](Int addr)
		{
			assert(verifyAddress(addr));
			return data[ addr.convertToint() ];
		}
	
		inline Word& top(int offset = 0)
		{
			if(SP < (Int)offset) throw Failure("Error in your code: not so many elements in heap");	
			return data[ SP.convertToint() - offset];
		}
	};
 }

#endif

