#ifndef CMA_INTERPRETER_DEFINED
#define CMA_INTERPRETER_DEFINED

#include <stdlib.h>
#include <vector>

#include "control.h"

namespace CMa
{
	/**
	 * An interpeter for the CMa that is parametrized with machine word class (Word), machine integer class (Int),
	 * heap implementation (Heap), free store imeplementation (FreeStore), and code segment implementation (Code).
	 *
	 * Assumptions:
	 *		-  The following values must fit in Int
	 *			- MAX_HEAP_ADDR + CMA_MAX_SP_DELTA
	 *			- MAX_CODE_ADDR + CMA_MAX_OP_CODE_ARG_COUNT
	 *			.
	 *		   where
	 *			- MAX_HEAP_ADDR = the maximum valid address in the heap
	 *			- MAX_CODE_ADDR = the maximum valid address in the code segment
	 *			- CMA_MAX_SP_DELTA, CMA_MAX_OP_CODE_ARG_COUNT = see file "op.h"
	 *
     *
	 *  Validity of C-machine instructions is tested at runtime. Proper errors are raised before
	 *  executing invalid instructions. Possible errors are:
	 *  - Calculation errors
	 *  - Stack ower/underflow
	 *  - memory accessing errors
	 *  - missing command wrong value of PC
	 *  - stack frame errors
	 *  - I/0 errors which are raised by Control
	 *
	 *  Implementation of the C-machine is realised by private functions look operations.h for further
	 *  details.
	 *
	 * @author OM/modified by SL
	 *
	 * STATUS: NOT TESTED
	 *
	 * TODO: Verify readoperation and functions in operations.h
	 */
	
	template<class Word, class Int, class Heap, class FreeStore, class Code> class Interpreter
	{
		public:
			Heap heap;
			Code code;

			FreeStore freeStore;

			// Registers
			Int PC;
			Int FP;
			/******************************Implementation*of*C-machine*operations*****************************/
			private:
				#include "operations.h"	
			/******************************************Public*methods*****************************************/
			public:
			
			Interpreter() : freeStore(heap){}

			void execute(Control<Word>* pControl)
			{
				try
				{
					// Internal variables
					Op::Code opCode;
					Int codeArg[CMA_MAX_OP_CODE_ARG_COUNT];

					Int newPC;
					Int newSP;

					// Initializing registers
					PC = (Int)0;
					FP = (Int)0;

					// Initializing subconstructions
					heap.init();
					freeStore.init();
					code.init();

					// Start of execution
					int tillBreak = pControl->onBreak();
				
					while (tillBreak != 0 )
					{
						//Both SP is correct, PC may be invalid
						readOperation(opCode, codeArg, newPC, newSP);
						//Now newSP is correct, newPC may be incorrect.
						//NewSP is calculated for every operation!!
						
						switch (opCode)
						{
							//Aritmetical operations
							case Op::MINUS: unary_minus(newSP); break;
							case Op::INC:   inc(newSP); break;
							case Op::DEC:   inc(newSP); break;
							case Op::MUL:   mul(newSP); break;
							case Op::ADD:   add(newSP); break;
							case Op::SUB:   sub(newSP); break;
							case Op::DIV:   div(newSP); break;
							case Op::MOD:   mod(newSP); break;
							//Logical operations
							case Op::NOT:   log_neg(newSP); break;
							case Op::AND:   log_and(newSP); break;
							case Op::OR:    log_or(newSP);  break;
							//Comparison operators
							case Op::EQ:    eq(newSP);  break;
							case Op::NEQ:   neq(newSP); break;
							case Op::LE:    le(newSP);  break;
							case Op::LEQ:   leq(newSP); break;
							case Op::GE:    ge(newSP);  break;
							case Op::GEQ:   geq(newSP); break;
							//Bitwise operators
							case Op::BIT_NOT: bin_complement(newSP); break;
							case Op::BIT_AND: bin_and(newSP); break;
							case Op::BIT_OR:  bin_or(newSP);  break;
							case Op::BIT_XOR: bin_xor(newSP); break;
							case Op::BIT_SHL: shift_left(newSP);  break;
							case Op::BIT_SHR: shift_right(newSP); break;
							//Various jumps
							case Op::JUMP:    jump(newSP, codeArg[0]);   break;
							case Op::JUMPZ:   jumpz(newSP,codeArg[0]);   break;
							case Op::JUMPI:   jumpi(newSP, codeArg[0]);  break;
							case Op::JUMPZS:  jumpzs(newSP, codeArg[0]); break;
							case Op::JUMPNZ:  jumpnzs(newSP,codeArg[0]); break;
							//Store and load
							case Op::LOADC:   push(newSP,(Word)codeArg[0]);  break;
							case Op::POP:     pop(newSP);    break;
							case Op::LOAD:    load(newSP);   break;
							case Op::LOADA:   loada(newSP, codeArg[0]);  break;
							case Op::STORE:   store(newSP);  break;
							case Op::STOREA:  storea(newSP, codeArg[0]); break;
							case Op::LOADRC:  loadrc(newSP, codeArg[0]); break;
							case Op::LOADR:   loadr(newSP, codeArg[0]);  break;
							case Op::STORER:  storer(newSP,codeArg[0]);  break;
							//Duplication
							case Op::DUP:     dup(newSP);  break;
							case Op::MOVE:    move(newSP, codeArg[0]); break;
							//Memory management
							case Op::NEW:     alloc_memory(newSP); break;
							case Op::FREE:    free_memory(newSP);  break;
							//Organisatory commands
							case Op::MARK:    mark(newSP); break;
							case Op::CALL:    call(newSP, codeArg[0]);  break;
							case Op::ENTER:   enter(newSP, codeArg[0]); break;
							case Op::ALLOC:   alloc_stack(newSP); break;
							case Op::RETURN:  return_value(newSP); break;
							case Op::HALT:
							    halt_machine(newSP,pControl);
							    return; //Nothing to do further
							    break;
							//I/O commands
							case Op::WRITEI:  write_int(newSP,pControl);  break;
							case Op::READI:   read_int(newSP,pControl);   break;
							case Op::WRITEC:  write_char(newSP,pControl,(Word)codeArg[0]); break;
							//This is THE END
							default: throw Failure("Error in completing instruction -- one operation isnt implemented(Sorry)!");
						}
					//Prepararing to fetch new instruction
					PC = newPC;
					tillBreak--;
					if (tillBreak == 0) tillBreak = pControl->onBreak();
					//End of while
				}
			}
			catch(Exception& error)
			{
				pControl->onException(&error);
			}
			//Closing up
			freeStore.deInit();
			heap.deInit();
			code.deInit();
			return;
		}

	protected:
		/**   \brief Command to identify next instruction and calculate newSP,newPC and possible code argument */
		/**  Reads in one instruction and its argument from the code.
		 *  Computes the new values of PC(newPC) and SP(newSP) FOR ALL INSTRUCTIONS.
		 *  Raises errors when instruction would corrupt state of stack or instruction is unknown:
		 *  - STACK_UNDERFLOW
		 *  - STACK_OVERFLOW
		 *  - WRONG_OP_CODE
		 *  - WRONG_ARGUMENT_ADDRESS
		 *
		 *  Hidden assumptions:
		 *  - SP is valid
		 *  - FP is valid
		 *  - codeArg has at least CMA_MAX_OP_CODE_ARG_COUNT elements
		 *
		 *  It ensures that newSP is valid under current EP, NP, FP and nothing more.
		 *
		 * @param opCode    operation code
		 * @param codeArg   operation arguments extracted from the code segment
		 *                  As most of the uses of code arguments require conversion to Int
		 *                  the result type is Int.
		 * @param newPC     the value of PC after executing this instruction
		 * @param newSP     the value of SP after executing this instruction
		 *
		 *
		 * @author OM/modified by SL
		 *
		 * TODO: Verify correctness.
		 */
		inline void readOperation(Op::Code& opCode,Int codeArg[],Int& newPC,Int& newSP) throw (Error)
		{
		    //THIS LOOKS UGLY AND IS UGLY!
		    //Check that PC is correct
			if(!code.verifyAddress(PC)) throw Error(Error::WRONG_CODE_ADDRESS);					

			// Load operation code
			opCode = code[PC].convertToCode();
			
			// Load information about the operation code
			int codeArgN;
			int stackArgN;
			int SPDelta;
			if (!op.getDescription(opCode, codeArgN, stackArgN, SPDelta))
				throw Error(Error::WRONG_OP_CODE);

			// Compute new PC
			newPC = PC + (Int)codeArgN + (Int)1;

			// Load code arguments
			assert(codeArgN <= CMA_MAX_OP_CODE_ARG_COUNT);
			if (codeArgN != 0)
			{
				if (!code.verifyInterval(PC, PC + (Int)codeArgN))
					throw Error(Error::WRONG_OP_ARGUMENT_ADDRESS);

				for(int i = 1; i <= codeArgN; i++)
					codeArg[i-1] = (Int)code[PC + (Int)i];
			}

			// Compute new SP
			if (SPDelta != 0)
			{
				newSP = heap.getSP() + (Int)SPDelta;
				//Big hack for ENTER
                //If program is translated then first ENTER  then SP is invalid
                if (PC!=(Int)0 || Op::ENTER!=opCode)
                {
				    if (SPDelta < 0 && !heap.verifySP(newSP)) throw Error(Error::STACK_UNDERFLOW);			
				    if (SPDelta > 0 && !heap.verifySP(newSP)) throw Error(Error::STACK_OVERFLOW);
			    }
			}
			// Compute newSP for operations: MOVE,ALLOC,RETURN,
			if(opCode==Op::MOVE)
			{
			    newSP = heap.getSP() + (Int)codeArg[0]-(Int)1;
			    if(!heap.verifySP(newSP)) throw Error(Error::STACK_OVERFLOW);
			}
			if(opCode==Op::ALLOC)
			{
				newSP = heap.getSP() + (Int)codeArg[0];
			    if(!heap.verifySP(newSP)) throw Error(Error::STACK_OVERFLOW);
			}
			if(opCode==Op::RETURN)
			{
			    newSP = FP-(Int)3;
			    if(!heap.verifySP(newSP)) throw Error(Error::STACK_UNDERFLOW);
			}
			//Shoud be correct now?

		}

	};
}

#endif

