#ifndef CMA_OPERATIONS_DEFINED
#define CMA_OPERATIONS_DEFINED
#ifndef CMA_INTERPRETER_DEFINED
    #error  operations.h must be included into the file interpreter.h
#endif  						
/**\file operations.h Implements C-machine instructions
 *  Impments the list of C-machine instructions inside class
 *  CMa::Interpreter< Word, Int, Heap, FreeStore, Code > Class
 *
 * Hidden assumptions:
 *
 * Assumptions on parameter class Int:
 *  It must posess following methods:
 *  - operator++(Word&)
 *  - operator--(Word&)
 *  - operator+ (Int&)
 *  - operator- (Int&)
 *  - operator--(Int&)
 *  - operator++(Int&)
 *  - operator< (Int&)
 *  - operator<=(Int&)
 *  - operator==(Int&)
 *  - operator!=(Int&)
 *  - Int(int&)
 *  - char(Int&)
 *
 * Assumptions on parameter class Word:
 *  It must posess following methods:
 *  - operator++(Word&)
 *  - operator--(Word&)
 *  - operator- (Word&) unary minus
 *  - operator*=(Word&)
 *  - operator+=(Word&)
 *  - operator-=(Word&)
 *  - operator/=(Word&)
 *  - operator%=(Word&)
 *  - operator<<(Word&)
 *  - operator>>(Word&)
 *  - operator||(Word&)
 *  - operator&&(Word&)
 *  - operator! () unary negotation
 *  - operator&=(Word&)
 *  - operator|=(Word&)
 *  - operator^=(Word&)
 *  - operator~ () unary complement
 *  - operator==(Word&)
 *  - operator!=(Word&)
 *  - operator< (Word&)
 *  - operator<=(Word&)
 *  - operator> (Word&)
 *  - operator>=(Word&)
 *  - operator= (Word&)
 *  - operator= (Int&)
 *  - Word(Int&)
 *  - Word(int&)
 *
 *  Visible (class) variables are:
 *  - Int PC, FP
 *  - Heap heap
 *	- Code code
 *  - Freestore freeStore
 *  - Control *pControl(onOutput(char*,Int&), onInput(Int&), onComlete())
 *
 *  Heap and Code methods verifyXXX have following semantics:
 *   - if everything is OK   they return TRUE
 *   - if something is WRONG they return FALSE
 *   - they dont change state of Heap/Code
 *
 *  For efficiency reasons we always use newSP as value for new SP value
 *  because it has been already calculated. So it is CRUCIAL that
 *  - SP     points to last stack record which is used
 *  - newSP  points to last stack record after succesful operation
 *    (the operation itselt doesn' have to be succesful)
 *
 *  We also assume that SP and newSP are valid stack addresses ie.
 *  - no STACK UNDERFLOW
 *  - no STACK_OVERFLOW
 *
 *  If operation has two or more arguments in stack then the last argument is on top,
 *  ie. op(x,y) = op(heap[SP-1],heap[SP]).
 *
 *  For unification the first argument of every operation is newSP, even it does't need it.
 *
 *  Operation mark raises SP to SP+4 we assume that this is still smallereq. EP ie.
 *  EP is correctly calculated.
 *
 *  In operation alloc_memory we assume that Freestore method malloc is clever enough
 *  to release memory when needed. If allocation doesn't succeed we return NULL=(Word)0.
 *
 *  Operations load, store, loada, storea, loadr, storer will raise Error when address is NULL
 *
 *  Operation loadq is RENAMED to push.
 *
 *  Operation move starts coping from top to bottom and newer produces uncorrect result.
 *
 *
 *  If the operation doesn't succeed the state of interpreter remains same as before
 *  executing operation.
 *
 *  List of all possible errors:
 *  - STACK_OVERFLOW
 *  - DIVISION_BY_ZERO
 *  - MODULUS_IS_ZERO
 *  - READ_FROM_NULL_POINTER
 *  - WRONG_HEAP_ADDRESS
 *  - WRITE_TO_NULL_POINTER
 *  - WRONG_JUMP_ADDRESS
 *  - WRONG_CALL_ADDRESS
 *  - WRONG_HEAP_INTERVALL
 *  - CALL_UNDERFLOW
 *  - CORRUPTED_FRAME
 *  - READ_FROM_STACK_FRAME
 *  - WRITE_TO_STACK_FRAME
 *
 *
 */					

/************************************Aritmetical*Operations**********************************************/
inline void inc(const Int& newSP)
{
	 ++heap.top();	
}

inline void dec(const Int& newSP)
{
	 --heap.top();
}

inline void unary_minus(const Int& newSP)
{
	heap.top() =- heap.top();
}

inline void mul(const Int& newSP)
{
	heap[newSP] *= heap.top();
	heap.setSP(newSP);
}

inline void add(const Int& newSP)
{
	heap[newSP] += heap.top();
	heap.setSP(newSP);
}

inline void sub(const Int& newSP)
{
	heap[newSP] -= heap.top();
	heap.setSP(newSP);
}

inline void div(const Int& newSP)
{
	heap[newSP] /= heap.top();
	heap.setSP(newSP);
}

inline void mod(const Int& newSP)
{
	heap[newSP] %= heap.top();
	heap.setSP(newSP);
}

/******************************************Logical*Operations********************************************/
inline void log_neg(const Int& newSP)
{
	heap.top() = !heap.top();
}

inline void log_and(const Int& newSP)
{
	heap[newSP] = heap[newSP] && heap.top();
	heap.setSP(newSP);
}

inline void log_or(const Int& newSP)
{
	heap[newSP] = heap[newSP] || heap.top();
	heap.setSP(newSP);
}

/************************************Comparison*Operators************************************************/
inline void eq(const Int& newSP)
{
	heap[newSP] = (Word)(heap[newSP] == heap.top());
    heap.setSP(newSP);
}

inline void neq(const Int& newSP)
{
	heap[newSP] = (Word)(heap[newSP] != heap.top());
	heap.setSP(newSP);
}

inline void le(const Int& newSP)
{
  heap[newSP] = (Word)(heap[newSP] < heap.top());
  heap.setSP(newSP);
}

inline void leq(const Int& newSP)
{
  heap[newSP] = (Word)(heap[newSP] <= heap.top());
  heap.setSP(newSP);
}

inline void ge(const Int& newSP)
{
  heap[newSP] = (Word)(heap[newSP] > heap.top());
  heap.setSP(newSP);
}

inline void geq(const Int& newSP)
{
  heap[newSP] = (Word)(heap[newSP] >= heap.top());
  heap.setSP(newSP);
}


/*******************************************Bit*Operators************************************************/
inline void bin_complement(const Int& newSP)
{
	heap.top() = ~heap.top();
}

inline void bin_and(const Int& newSP)
{
	heap[newSP] &= heap.top();
	heap.setSP(newSP);
}

inline void bin_or(const Int& newSP)
{
	heap[newSP] |= heap.top();
	heap.setSP(newSP);
}

inline void bin_xor(const Int& newSP)
{
	heap[newSP] ^= heap.top();
	heap.setSP(newSP);
}

inline void shift_left(const Int& newSP)
{
	//shifts value x  y bits to left
	heap[newSP]<<=heap.top();
	heap.setSP(newSP);
}

inline void shift_right(const Int& newSP)
{
	//shifts value x y bits to right
    heap[newSP]>>=heap.top();
    heap.setSP(newSP);
}

/*****************************************Controll*commands**********************************************/
inline void jump(const Int& newSP, const Int& addr)
{
	if(!code.verifyAddress(addr)) throw Error(Error::WRONG_JUMP_ADDRESS);
	PC = addr;
}

inline void jumpz(const Int& newSP, const Int& addr)
{
	if(heap.top()==(Word)0)
	{
		if(!code.verifyAddress(addr)) throw Error(Error::WRONG_JUMP_ADDRESS);
		PC = addr;
	}
	heap.setSP(newSP);
}

inline void jumpi(const Int& newSP, const Int& addr)
{
	if(!code.verifyAddress(addr+(Int)heap.top())) throw Error(Error::WRONG_JUMP_ADDRESS);
	PC = addr + (Int)heap.top();
	heap.setSP(newSP);
}

inline void jumpzs(const Int& newSP, const Int& addr)
{
	if(heap.top()==(Word)0)
	{
		if(!code.verifyAddress(addr)) throw Error(Error::WRONG_JUMP_ADDRESS);
		PC = addr;
	}
}

inline void jumpnzs(const Int& newSP, const Int& addr)
{
	if(heap.top()!=(Word)0)
	{
		if(!code.verifyAddress(addr)) throw Error(Error::WRONG_JUMP_ADDRESS);
		PC = addr;
	}
}

/*******************************Storage*and*retrievial*commands******************************************/
inline void push(const Int& newSP, const Word& value)
{
	heap.setSP(newSP);		
	heap[newSP] = value;
}

inline void  pop(const Int& newSP)
{
	heap.setSP(newSP);
}

inline void load(const Int& newSP)
{
    if(heap[newSP]==(Word)0)  throw Error(Error::READ_FROM_NULL_POINTER);
	if(!heap.verifyAddress((Int)heap[newSP])) throw Error(Error::WRONG_HEAP_ADDRESS);
	heap[newSP] = heap[(Int)heap[newSP]];
}

inline void store(const Int& newSP)
{
    if(heap.top()==(Word)0)  throw  Error(Error::WRITE_TO_NULL_POINTER);
	if(!heap.verifyAddress((Int)heap.top())) throw Error(Error::WRONG_HEAP_ADDRESS);
	heap[(Int)heap.top()] = heap[newSP];
	heap.setSP(newSP);
}

inline void loada(const Int& newSP, const Int& addr)
{
    if(addr==(Int)0) throw Error(Error::READ_FROM_NULL_POINTER);
	if(!heap.verifyAddress(addr)) throw Error(Error::WRONG_HEAP_ADDRESS);
	heap.setSP(newSP);	
	heap[newSP] = heap[addr];
}

inline void storea(const Int& newSP, const Int& addr)
{
    if(addr==(Int)0) throw Error(Error::WRITE_TO_NULL_POINTER);
	if(!heap.verifyAddress(addr)) throw Error(Error::WRONG_HEAP_ADDRESS);
	heap[addr] = heap.top();
	heap.setSP(newSP);
}

inline void loadrc(const Int& newSP,const Int& value)
{
	heap.setSP(newSP);		
	heap[newSP] =(Word)(FP + value);
}

inline void loadr(const Int& newSP, const Int& addr)
{
    if(addr==(Int)0) throw Error(Error::READ_FROM_NULL_POINTER);
    if(addr==(Int)-1 || addr==(Int)-2) throw Error(Error::READ_FROM_STACK_FRAME);
	if(!heap.verifyAddress(FP+addr)) throw Error(Error::WRONG_HEAP_ADDRESS);	
    heap.setSP(newSP);
    heap[newSP] = heap[FP+addr];
}

inline void storer(const Int& newSP, const Int& addr)
{
    if(addr==(Int)0) throw Error(Error::WRITE_TO_NULL_POINTER);
    if(addr==(Int)-1 || addr==(Int)-2) throw Error(Error::WRITE_TO_STACK_FRAME);
	if(!heap.verifyAddress(FP+addr)) throw Error(Error::WRONG_HEAP_ADDRESS);	
    heap[FP+addr] = heap.top();
    heap.setSP(newSP);	
}

inline void dup(const Int& newSP)
{
	heap.setSP(newSP);
	heap[newSP] = heap[newSP-(Int)1];
}

inline void move(const Int& newSP, const Int& number)
{
    Int oldSP = heap.getSP();
    Int addr = (Int)heap.top();
	if(!heap.verifyInterval(addr, addr+number-(Int)1)) throw Error(Error::WRONG_HEAP_INTERVALL);
	//Coping from bottom to top. If addr-number-1==SP then overwrite error
	for(Int i=number-(Int)1;(Int)0<=i;--i) heap[oldSP+i] = heap[addr+i];
}

/****************************************Memory*(de)allocation*******************************************/
inline void alloc_memory(const Int& newSP)
{
	if(heap.top()==(Word)0) return;
	//It is its own business how freestore allocates memory
	heap.top() = freeStore.malloc((Int)heap.top());
}

inline void free_memory(const Int& newSP)
{
	heap.setSP(newSP);	
}

/***************************************Organisatory*Commands********************************************/
inline void mark(const Int& newSP)
{
	heap.setSP(newSP);
	heap.top(1) = (Word)FP;
	heap.top(2) = (Word)heap.getEP();
}

inline void call(const Int& newSP, const Int& value)
{
	if(newSP<= value) throw Error(Error::CALL_UNDERFLOW);
	if(!code.verifyAddress((Int)heap.top())) throw Error(Error::WRONG_CALL_ADDRESS);
	FP = newSP - value;	
	heap[FP] = (Word)PC;
	PC = (Int) heap.top();
	heap.setSP(newSP);	
}

inline void enter(const Int& newSP, const Int& value)
{
	if(!heap.verifyEP(value+heap.getSP())) throw Error(Error::STACK_OVERFLOW);
	heap.setEP(value+heap.getSP());
}

inline void alloc_stack(const Int& newSP)
{
	heap.setSP(newSP);
}

inline void return_value(const Int& newSP)
{
	if(!heap.verifyAddress(FP))              throw Error(Error::CORRUPTED_FRAME);   	
	if(!code.verifyAddress((Int)heap[FP]))   throw Error(Error::CORRUPTED_FRAME);
	if(!heap.verifyAddress(FP-(Int)2))       throw Error(Error::CORRUPTED_FRAME);   		
	if(!heap.verifyAddress(FP-(Int)1))       throw Error(Error::CORRUPTED_FRAME);  	
	if(!heap.verifyEP((Int)heap[FP-(Int)2])) throw Error(Error::STACK_OVERFLOW);
	if(!heap.verifyFP((Int)heap[FP-(Int)1],FP-(Int)3)) throw Error(Error::CORRUPTED_FRAME);
	PC = (Int)heap[FP];
	FP = (Int)heap[FP-(Int)1];	
	heap.setSP(FP-(Int)3);
}

void halt_machine(const Int& newSP, Control<Word>* pControl)
{

	if(heap.getSP()!=(Int)1) throw Error("Wrong stack state on HALT");	

    freeStore.verifyHalt();
	pControl->onComplete(heap[(Int)1]);
	freeStore.deInit();
	heap.deInit();
	code.deInit();	
}

/********************************************I/O*operations**********************************************/
inline void write_int(const Int& newSP, Control<Word>* pControl)
{
	pControl->onOutput_int(heap.top());
	heap.setSP(newSP);
}

void read_int(const Int& newSP, Control<Word>* pControl)
{
	heap.setSP(newSP);
	heap[newSP] = pControl->onInput();
}

void write_char(const Int& newSP, Control<Word>* pControl, const Word& outchar)
{
	pControl->onOutput_char(outchar);
}

#endif